/* $Id: fx100.cpp,v 1.18 2002/07/15 05:19:23 chrisf Exp $ */

/*
Copyright (C) 2002 Chris Faherty

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <qglobal.h>
#include <qpe/qcopenvelope_qws.h>
#include "fx100.h"

#define KB_LEFT 1
#define KB_RIGHT 2
#define KB_UP 3
#define KB_DOWN 4
#define KB_LSHIFT 5
#define KB_RSHIFT 6
#define KB_FUNC 7
#define KB_BACKSPACE 8
#define KB_TAB 9  /* skipped bs */
#define KB_COMMAND 10
#define KB_SHORTCUT 11
#define KB_F1 12
#define KB_ENTER 13
#define KB_F2 14 /* skipped cr */
#define KB_F3 15
#define KB_F4 16
#define KB_F5 17
#define KB_F6 18
#define KB_F7 19
#define KB_F8 20
#define KB_F9 21
#define KB_F10 22
#define KB_CTL 23
#define KB_CAPLOCK 24
#define KB_WIN 25
#define KB_HOME 26
#define KB_END 27
#define KB_MENU 28
#define KB_PGUP 29
#define KB_PGDN 30
#define KB_INSERT 31
#define KB_DELETE 127

const unsigned char flexis_lut[]={
/* 0x00 */ 'a', 's', 'd', 'f', 'h', 'g', 'z', 'x',
/* 0x08 */ 'c', 'v', KB_WIN, 'b', 'q', 'w', 'e', 'r',
/* 0x10 */ 'y', 't', '1', '2', '3', '4', '6', '5',
/* 0x18 */ '=', '9', '7', '-', '8', '0', ']', 'o',
/* 0x20 */ 'u', '[', 'i', 'p', KB_ENTER, 'l', 'j', '\'',
/* 0x28 */ 'k', ';', '\\', ',', '/', 'n', 'm', '.',
/* 0x30 */ KB_TAB, ' ', '`', KB_BACKSPACE, 0, 0, 0, KB_SHORTCUT,
/* 0x38 */ KB_LSHIFT, KB_CAPLOCK, KB_FUNC, KB_CTL, 0, KB_COMMAND, 0, KB_RSHIFT,
/* 0x40 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x48 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x50 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x58 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x60 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x68 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x70 */ 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x78 */ 0, 0, KB_DELETE, KB_LEFT, KB_RIGHT, KB_DOWN, KB_UP, 0
};

const unsigned char shift_nums[]={
    ')', '!', '@', '#', '$', '%', '^', '&', '*', '('
};

FX100::FX100(QObject *parent=0, const char *name=0) :
    QObject(parent, name),
    socketRead(NULL),
    fd(-1),
    shift(false),
    capslock(false),
    ctrl(false),
    alt(false),
    func(false)
{
}

FX100::~FX100()
{
    close_serial();
}

bool FX100::enable(void)
{
    return open_serial();
}

bool FX100::disable(void)
{
    return close_serial();
}

bool FX100::open_serial(void)
{
    fd = open("/dev/ttyS0", O_RDONLY|O_NOCTTY|O_NONBLOCK);
    if (fd < 0) return false;
    struct termios t;
    /* set the baud rate and make it raw */
    tcgetattr(fd, &t);
    t.c_cflag = CS8|CREAD|HUPCL|CLOCAL;
    t.c_iflag = IGNBRK|IGNPAR;
    t.c_oflag = 0;
    t.c_lflag = 0;
    t.c_cc[VMIN ] = 1;
    t.c_cc[VTIME] = 0;
    cfsetispeed(&t, B9600);
    cfsetospeed(&t, B9600);
	tcsetattr(fd, TCSANOW, &t);
    /* raise DTR, RTS */
    int i=0;
    i |= TIOCM_DTR|TIOCM_RTS;
    ioctl(fd, TIOCMSET, &i);
    if ((socketRead = new QSocketNotifier(fd, QSocketNotifier::Read))) {
        connect(socketRead, SIGNAL(activated(int)),
        this, SLOT(dataReceived(int)));
    }
    return true;
}

bool FX100::close_serial(void)
{
    if (fd < 0) return true;
    if (socketRead) {
        delete socketRead;
        socketRead = NULL;
    }
    return ((close(fd) == 0) && (fd = -1));
}

void FX100::dataReceived(int s)
{
    uchar raw, cook;
    while (read(s, &raw, 1) > 0) {
        cook = raw&0x7f;
        if (cook >= sizeof(flexis_lut)) continue;
        ushort uc=flexis_lut[cook];
        ushort qc=0;
        switch (uc) {
        case KB_ENTER: qc = Qt::Key_Return; break;
        case KB_BACKSPACE: qc = Qt::Key_Backspace; break;
        case KB_UP: qc = Qt::Key_Up; break;
        case KB_DOWN: qc = Qt::Key_Down; break;
        case KB_LEFT: qc = Qt::Key_Left; break;
        case KB_RIGHT: qc = Qt::Key_Right; break;
        case KB_TAB: qc = Qt::Key_Tab; break;
        case KB_WIN: qc = Qt::Key_Escape; uc = 27; break;
        case KB_DELETE: qc = Qt::Key_Delete; break;
        case KB_LSHIFT:
        case KB_RSHIFT:
            shift = (bool)(raw&0x80);
            qc = Qt::Key_Shift;
            uc = 0;
            break;
        case KB_CAPLOCK:
            if (raw&0x80) capslock = !capslock; /* toggle */
            qc = Qt::Key_Shift;
            uc = 0;
            break;
        case KB_CTL:
            ctrl = (bool)(raw&0x80);
            qc = Qt::Key_Control;
            uc = 0;
            break;
        case KB_SHORTCUT:
            alt = (bool)(raw&0x80);
            qc = Qt::Key_Alt;
            uc = 0;
            break; /* no Alt on fx100 */
        case KB_FUNC:
            func = (bool)(raw&0x80);
            qc = 0;
            uc = 0;
            break;
        }
        if (!(raw&0x80)) continue; /* high-bit off is break code */

        /* launches */
        if (func) {
            switch (uc) {
            case 'a':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("filebrowser");
                break;
                }
            case '1':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("addressbook");
                break;
                }
            case '2':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("textedit");
                break;
                }
            case '3':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("datebook");
                break;
                }
            case '4':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("todolist");
                break;
                }
            case 'f':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("showimg");
                break;
                }
            case '=':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("light-and-power");
                break;
                }
            case 'k':
                {
                QCopEnvelope e("QPE/System", "execute(QString)");
                e << QString("embeddedkonsole");
                break;
                }
            }
            continue;
        }

        /* make it into control chars if needed */
        if (ctrl) {
            if ((uc >= 'a') && (uc <= 'z')) uc = uc-'a'+1;
            if ((uc >= '[') && (uc <= '_')) uc = uc-'A';
        }
        /* these chars only if shift is active, not capslock */
        if (shift) {
            /* the numbers */
            if ((uc >= '0') && (uc <= '9')) {
                uc = shift_nums[uc-'0'];
            }
            /* other chars */
            switch (uc) {
            case '-': uc = '_'; break;
            case '=': uc = '+'; break;
            case '\\': uc = '|'; break;
            case ';': uc = ':'; break;
            case '\'': uc = '"'; break;
            case ',': uc = '<'; break;
            case '.': uc = '>'; break;
            case '/': uc = '?'; break;
            case '[': uc = '{'; break;
            case ']': uc = '}'; break;
            case '`': uc = '~'; break;
            }
        }
        QChar c(uc); /* for the unicode stuff */
        /* I still have to manually shift characters, even with mod */
        if (shift || capslock) c = c.upper(); /* make uppercase */
        /*
         * I'm not so sure how this stuff works, but some applications
         * are picky about chars.
         */
        if (qc == 0) {
            qc = uc;
        }
        /* make the modifiers parameter */
        ushort mod=
            ((shift || capslock) ? Qt::ShiftButton : 0)|
            (ctrl ? Qt::ControlButton : 0)|
            (alt ? Qt::AltButton : 0);
        emit keypress(c.unicode(), qc, mod);
    }
}
